/***************************************************************************
    copyright            : (C) 2005 by Brian Kerrick Nickel
    email                : brian.nickel@gmail.com
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

#include "TagData.h"


#include <sys/stat.h>


/* LOAD UP THE FILE'S TAG
 * Search each file type individually so we can get individual tags for
 * custom tagging.
 */
extern "C" TagData * tag_data_load (char * filename, int scanaudio)
{
   TagData * data;
   TagLib::String s = filename;
   TagLib::AudioProperties::ReadStyle style = scanaudio ? TagLib::AudioProperties::Average : TagLib::AudioProperties::Fast;
   int mtime;
   struct stat buf;
   if (!stat (filename, &buf))
      mtime = (int) buf.st_mtime;

   /* FIXME: Using filename to find media type. GStreamer probe instead?
    */
   if(s.size() > 4) {
      if(s.substr(s.size() - 4, 4).upper() == ".OGG")
      {
         TagLib::Vorbis::File * f = new TagLib::Vorbis::File(filename, style);
         if (! f->isValid ())
            return NULL;
         data = (TagData*) malloc (sizeof (TagData));
         data->file = new TagLib::FileRef (f);
         data->id3v2 = NULL;
         data->id3v1 = NULL;
         data->ape   = NULL;
         data->xiph  = f->tag ();
         data->mime  = "application/ogg";
         data->mtime = mtime;
         return data;
      }
      if(s.substr(s.size() - 4, 4).upper() == ".MP3")
      {
         TagLib::MPEG::File * f = new TagLib::MPEG::File(filename, style);
         if (! f->isValid ())
            return NULL;
         data = (TagData*) malloc (sizeof (TagData));
         data->file = new TagLib::FileRef (f);
         data->id3v2 = f->ID3v2Tag ();
         data->id3v1 = f->ID3v1Tag ();
         data->ape   = f->APETag ();
         data->xiph  = NULL;
         data->mime  = "audio/mpeg";
         data->mtime = mtime;
         return data;
      }
      if(s.substr(s.size() - 5, 5).upper() == ".FLAC")
      {
         TagLib::FLAC::File * f = new TagLib::FLAC::File(filename, style);
         if (! f->isValid ())
            return NULL;
         data = (TagData*) malloc (sizeof (TagData));
         data->file = new TagLib::FileRef (f);
         data->id3v2 = f->ID3v2Tag ();
         data->id3v1 = f->ID3v1Tag ();
         data->ape   = NULL;
         data->xiph  = f->xiphComment ();
         data->mime  = "audio/x-flac";
         data->mtime = mtime;
         return data;
      }
      if(s.substr(s.size() - 4, 4).upper() == ".MPC")
      {
         TagLib::MPC::File * f = new TagLib::MPC::File(filename, style);
         if (! f->isValid ())
            return NULL;
         data = (TagData*) malloc (sizeof (TagData));
         data->file = new TagLib::FileRef (f);
         data->id3v2 = NULL;
         data->id3v1 = f->ID3v1Tag ();
         data->ape   = f->APETag ();
         data->xiph  = NULL;
         data->mime  = "audio/x-musepack";
         data->mtime = mtime;
         return data;
      }
   }
   return NULL;
}


/* SAVE CHANGES TO TAG
 */
extern "C" bool tag_data_save (TagData * data)
{
   return data->file->save ();
}


/* FREE CHANGES
 * FIXME: Is all memory freed, check for leaks.
 */
extern "C" void tag_data_free (TagData * data)
{
   if (data == NULL)
      return;
   delete data->file;
   free (data);
}

/* SET ID3V2 FRAME ENCODING
 */
void tag_data_set_id3v2_encoding (TagData * data, const char * frame)
{

   TagLib::ID3v2::FrameList l;
   TagLib::ID3v2::TextIdentificationFrame * f;
   if (!data->id3v2)
      return;
   l = data->id3v2->frameListMap()[frame];
   if (l.isEmpty ())
      return;
   f = dynamic_cast<TagLib::ID3v2::TextIdentificationFrame*> (l.front ());
   f->setTextEncoding (TagLib::String::UTF8);

}




/* GET/SET SONG TITLE
 */
extern "C" const char * tag_data_get_title (TagData * data)
{
   return ::strdup(data->file->tag ()->title ().toCString (1));
}
extern "C" void tag_data_set_title (TagData * data, char * val)
{
   data->file->tag ()->setTitle (TagLib::String(val, TagLib::String::UTF8));
   tag_data_set_id3v2_encoding (data, "TIT2");
   free (val);
}


/* GET/SET ALBUM
 */
extern "C" const char * tag_data_get_album (TagData * data)
{
   return ::strdup(data->file->tag ()->album ().toCString (1));
}
extern "C" void tag_data_set_album (TagData * data, char * val)
{
   data->file->tag ()->setAlbum (TagLib::String(val, TagLib::String::UTF8));
   tag_data_set_id3v2_encoding (data, "TALB");
   free (val);
}


/* GET/SET DISC NAME
 */
extern "C" char * tag_data_get_disc_name (TagData * data)
{
   return tag_data_get_frame_string (data, "DISCNAME", "TSST");
}
extern "C" void tag_data_set_disc_name (TagData * data, char * str)
{
   tag_data_set_frame_string (data, "DISCNAME", "TSST", str);
   free (str);
}


/* GET/SET COMMENT
 */
extern "C" const char * tag_data_get_comment (TagData * data)
{
   return ::strdup(data->file->tag ()->comment ().toCString (1));
}
extern "C" void tag_data_set_comment (TagData * data, char * val)
{
   data->file->tag ()->setComment (TagLib::String(val, TagLib::String::UTF8));
   if (data->id3v2 &&!data->id3v2->frameListMap()["COMM"].isEmpty ())
      (dynamic_cast<TagLib::ID3v2::CommentsFrame*> (data->id3v2->frameListMap()["COMM"].front ()))->setTextEncoding (TagLib::String::UTF8);
   free (val);
}


/* GET/SET ARTISTS
 */
extern "C" StringList * tag_data_get_artists (TagData * data)
{
   StringList * sl = tag_data_get_frame_strings (data, "ARTIST", "TPE1");
   return sl ? sl : string_list_from_string (data->file->tag ()->artist ());
}
extern "C" void tag_data_set_artists (TagData * data, StringList * sl)
{
   data->file->tag ()->setArtist (string_list_to_string (sl));
   tag_data_set_frame_strings (data, "ARTIST", "TPE1", sl);
   string_list_free (sl);
}


/* GET/SET PERFORMERS
 */
extern "C" StringList * tag_data_get_performers (TagData * data)
{
   return tag_data_get_frame_strings (data, "PERFORMER", "TPE2");
}
extern "C" void tag_data_set_performers (TagData * data, StringList * sl)
{
   tag_data_set_frame_strings (data, "PERFORMER", "TPE2", sl);
   string_list_free (sl);
}


/* GET/SET GENRES
 */
extern "C" StringList * tag_data_get_genres (TagData * data)
{
   StringList * sl = tag_data_get_frame_strings (data, "GENRE", "TCON");
   return sl ? sl : string_list_from_string (data->file->tag ()->genre ());
}
extern "C" void tag_data_set_genres (TagData * data, StringList * sl)
{
   data->file->tag ()->setGenre (string_list_to_string (sl));
   tag_data_set_frame_strings (data, "GENRE", "TCON", sl);
   string_list_free (sl);
}


/* GET/SET MOOD
 */
extern "C" StringList * tag_data_get_moods (TagData * data)
{
   return tag_data_get_frame_strings (data, "MOOD", "TMOO");
}
extern "C" void tag_data_set_moods (TagData * data, StringList * sl)
{
   tag_data_set_frame_strings (data, "MOOD", "TMOO", sl);
   string_list_free (sl);
}


/* GET/SET YEAR
 */
extern "C" uint tag_data_get_year (TagData * data)
{
   return data->file->tag ()->year ();
}
extern "C" void tag_data_set_year (TagData * data, uint val)
{
   data->file->tag ()->setYear (val);
}


/* GET/SET SONG TRACK NUMBER/TOTAL TRACKS
 */
extern "C" uint tag_data_get_track (TagData * data)
{
   return data->file->tag ()->track ();
}
extern "C" uint tag_data_get_track_count (TagData * data)
{
   uint num;
   
   num = tag_data_get_xiph_uint (data, "TRACKTOTAL");
   if (num) return num;
   num = tag_data_get_ape_uint (data, "TRACKTOTAL");
   if (num) return num;
   
   TagLib::String str = tag_data_get_id3v2_tstring (data, "TRCK");
   if (str.find ("/") > -1)
      return (uint) str.substr (str.find ("/")+1).toInt ();
   return 0;
}
extern "C" void tag_data_set_track (TagData * data, uint track, uint track_count)
{
   data->file->tag ()->setTrack (track);
   tag_data_set_xiph_uint (data, "TRACKTOTAL", track_count);
   tag_data_set_ape_uint (data, "TRACKTOTAL", track_count);
   if (track && track_count)
      tag_data_set_id3v2_tstring (data, "TRCK", TagLib::String::number(track) + "/" + TagLib::String::number(track_count), 0);
}


/* GET/SET DISC NUMBER
 * FIXME: Include number of discs in set? #/#
 */
extern "C" uint tag_data_get_disc_number (TagData * data)
{
   return tag_data_get_frame_uint (data, "DISCNUMBER", "TPOS");
}
extern "C" uint tag_data_get_disc_count (TagData * data)
{
   uint num;
   
   num = tag_data_get_xiph_uint (data, "DISCTOTAL");
   if (num) return num;
   num = tag_data_get_ape_uint (data, "DISCTOTAL");
   if (num) return num;
   
   TagLib::String str = tag_data_get_id3v2_tstring (data, "TPOS");
   if (str.find ("/") > -1)
      return (uint) str.substr (str.find ("/")+1).toInt ();
   return 0;
}
extern "C" void tag_data_set_disc_number (TagData * data, uint disc, uint disc_count)
{
   tag_data_set_xiph_uint (data, "DISCNUMBER", disc);
   tag_data_set_ape_uint  (data, "DISCNUMBER", disc);
   tag_data_set_xiph_uint (data, "DISCTOTAL", disc_count);
   tag_data_set_ape_uint  (data, "DISCTOTAL", disc_count);
   if (disc && disc_count)
      tag_data_set_id3v2_tstring (data, "TPOS", TagLib::String::number(disc) + "/" + TagLib::String::number(disc_count), 0);
   else
      tag_data_set_id3v2_uint (data, "TPOS", disc);
}


/* GET SONG LENGTH
 */
extern "C" uint tag_data_get_duration (TagData * data)
{
   return data->file->audioProperties()->length();
}

/* GET/SET REPLAY GAIN
 */
extern "C" float tag_data_get_gain (TagData * data)
{
   TagLib::String str;
   if (!((str = tag_data_get_xiph_tstring (data, "REPLAYGAIN_TRACK_GAIN")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_xiph_tstring (data, "REPLAYGAIN_ALBUM_GAIN")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_xiph_tstring (data, "RG_AUDIOPHILE")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_xiph_tstring (data, "RG_RADIO")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "REPLAYGAIN_TRACK_GAIN")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "REPLAYGAIN_ALBUM_GAIN")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "RG_AUDIOPHILE")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "RG_RADIO")) == ""))
      return atof (str.toCString (0));
   
   float f = 0.0;
   if (data->id3v2 && !data->id3v2->frameListMap()["RVA2"].isEmpty())
   {
      TagLib::ID3v2::FrameList::ConstIterator it;
      TagLib::ID3v2::FrameList l = data->id3v2->frameListMap()["RVA2"];
      for(it=l.begin(); it!=l.end(); ++it)
      {
         TagLib::ID3v2::RelativeVolumeFrame * rva = dynamic_cast<TagLib::ID3v2::RelativeVolumeFrame*>(*it);
         if (rva->volumeAdjustmentIndex (TagLib::ID3v2::RelativeVolumeFrame::MasterVolume))
         {
            if (f < 0.001 && f > - 0.001)
               f = rva->volumeAdjustment (TagLib::ID3v2::RelativeVolumeFrame::MasterVolume);
            if (rva->toString() == "Muine")
            {
               f = rva->volumeAdjustment (TagLib::ID3v2::RelativeVolumeFrame::MasterVolume);
               break;
            }
         }
      }
   }
   return f;
}
extern "C" void tag_data_set_gain (TagData * data, float val)
{
   if (val < -64.0) val = -64.0;
   if (val > 64.0)  val = 64.0;
   
   char strval [32];
   sprintf (strval, "%f", (double) val);
   
   tag_data_set_xiph_tstring (data, "REPLAYGAIN_TRACK_GAIN", strval);
   tag_data_set_ape_tstring (data, "REPLAYGAIN_TRACK_GAIN", strval);
   if (data->id3v2)
   {
      TagLib::ID3v2::RelativeVolumeFrame * rva = NULL;
      TagLib::ID3v2::FrameList::ConstIterator it;
      TagLib::ID3v2::FrameList l = data->id3v2->frameListMap()["RVA2"];
      for(it=l.begin(); it!=l.end(); ++it)
      {
         TagLib::ID3v2::RelativeVolumeFrame * tmprva = dynamic_cast<TagLib::ID3v2::RelativeVolumeFrame*>(*it);
         if (rva == NULL) rva = tmprva;
         if (tmprva->toString () == "Muine")
         {
            rva = tmprva;
            break;
         }
      }
      if (rva == NULL)
      {
         rva = new TagLib::ID3v2::RelativeVolumeFrame(TagLib::ByteVector("RVA2\0\0\0\x06\0\0Muine\0", (uint)16));
         data->id3v2->addFrame (rva);
      }
      rva->setVolumeAdjustmentIndex (short(val * float(512)), TagLib::ID3v2::RelativeVolumeFrame::MasterVolume);
   }
}

/* GET PEAK
 */
extern "C" float tag_data_get_peak (TagData * data)
{
   TagLib::String str;
   if (!((str = tag_data_get_xiph_tstring (data, "REPLAYGAIN_TRACK_PEAK")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_xiph_tstring (data, "REPLAYGAIN_ALBUM_PEAK")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_xiph_tstring (data, "RG_PEAK")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "REPLAYGAIN_TRACK_PEAK")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "REPLAYGAIN_ALBUM_PEAK")) == ""))
      return atof (str.toCString (0));
   if (!((str = tag_data_get_ape_tstring (data, "RG_PEAK")) == ""))
      return atof (str.toCString (0));
   
   /* FIXME: I've got not clue about ID3 peak. */

   return 0.0;
}


/* GET ALBUM COVER
 */
extern "C" const char * tag_data_get_cover (TagData * data, uint * len)
{
   TagLib::ID3v2::AttachedPictureFrame * pic = NULL;
   if (data->id3v2 && !data->id3v2->frameListMap()["APIC"].isEmpty())
   {
      TagLib::ID3v2::FrameList::ConstIterator it;
      TagLib::ID3v2::FrameList l = data->id3v2->frameListMap()["APIC"];
      for(it=l.begin(); it!=l.end(); ++it)
      {
         TagLib::ID3v2::AttachedPictureFrame * tmppic = dynamic_cast<TagLib::ID3v2::AttachedPictureFrame*>(*it);
         if (pic == NULL) pic = tmppic;
         if (tmppic->type () == TagLib::ID3v2::AttachedPictureFrame::FrontCover) 
         {
            pic = tmppic;
            break;
         }
      }
   }
   if (pic == NULL)
   {
      *len = 0;
      return NULL;
   }
   *len = pic->picture ().size ();
   const char * out = (const char *) malloc (*len);
   memcpy ((void *) out, pic->picture ().data(), *len);
   return out;
}


/* GET MIME TYPE
 */
extern "C" const char * tag_data_get_mime (TagData * data)
{
   return ::strdup(data->mime);
}


/* GET MTIME
 */
extern "C" int tag_data_get_mtime (TagData * data)
{
   return data->mtime;
}
